#!/usr/bin/env python
# coding: utf-8
"""
  chanjo.utils
  ~~~~~~~~~~~~~

  A number of utilities used elsewhere in Chanjo.

  :copyright: (c) 2013 by Robin Andeer
  :license: MIT, see LICENSE for more details
"""

class Interval(object):
  """
  (Genomic) Interval object. Note that both start and end are 0-based. If you
  don't provide an ``end`` parameter, a single position interval with be
  created.

  :param int start: The 0-based start position of the interval
  :param int end: The 0-based end position of the interval (optional)
  :param number value: A number representing the read depth across the
                       interval (optional)
  :param str chrom: The ID of the chromosome the interval belongs to
                    (optional)
  """
  def __init__(self, start, end=None, value=None, chrom=None):
    super(Interval, self).__init__()
    self.start = start
    # A single position interval can be set up by omitting end argument
    self.end = end or start
    self.value = value
    self.chrom = chrom

  def __len__(self):
    # We are counting the number of positions in the interval
    return (self.end - self.start) + 1

  def __str__(self):
    # This is the BED standard definition of an interval
    return "({start}, {end})".format(start=self.start, end=self.end)

  def __eq__(self, other):
    # This compares Interval instances by matches values
    return (self.start == other.start and self.end == other.end and
            self.value == other.value and self.chrom == other.chrom)



#!/usr/bin/env python
# coding: utf-8
"""
  chanjo.sql
  ~~~~~~~~~~~~~

  The default :class:`ElementAdapter` that ships with Chanjo. Provides an
  interface to a SQLite database using the SQLAlchemy ORM.

  This module also defines the ORM objects and adds nessesary methods.

  :copyright: (c) 2013 by Robin Andeer
  :license: MIT, see LICENSE for more details
"""
from elemental.core import Base, ElementalDB
import sqlalchemy as sa
from sqlalchemy.orm import relationship, backref


# =========================================================
#   Data ORM classes
# ---------------------------------------------------------
class GeneData(Base):
  """docstring for GeneData"""
  __tablename__ = "GeneData"

  id = sa.Column(sa.Integer, primary_key=True, autoincrement=True)
  coverage = sa.Column(sa.Float)
  completeness = sa.Column(sa.Float)

  # These column maps coverage/completeness to an individual+group
  sample_id = sa.Column(sa.String)
  group_id = sa.Column(sa.Integer)

  # Genetic relationship
  element_id = sa.Column(sa.String, sa.ForeignKey("Gene.id"))
  element = relationship("Gene", backref=backref("data"))

  def __init__(self, element_id, sample_id=None, group_id=None,
               coverage=None, completeness=None):
    super(GeneData, self).__init__()
    
    self.element_id = element_id
    self.sample_id = sample_id
    self.group_id = group_id
    self.coverage = coverage
    self.completeness = completeness


class TranscriptData(Base):
  """docstring for TranscriptData"""
  __tablename__ = "TranscriptData"

  id = sa.Column(sa.Integer, primary_key=True, autoincrement=True)
  coverage = sa.Column(sa.Float)
  completeness = sa.Column(sa.Float)

  # These column maps coverage/completeness to an individual+group
  sample_id = sa.Column(sa.String)
  group_id = sa.Column(sa.Integer)

  # Genetic relationship
  element_id = sa.Column(sa.String, sa.ForeignKey("Transcript.id"))
  element = relationship("Transcript", backref=backref("data"))

  def __init__(self, element_id, sample_id=None, group_id=None,
               coverage=None, completeness=None):
    super(TranscriptData, self).__init__()
    
    self.element_id = element_id
    self.sample_id = sample_id
    self.group_id = group_id
    self.coverage = coverage
    self.completeness = completeness


class ExonData(Base):
  """docstring for ExonData"""
  __tablename__ = "ExonData"

  id = sa.Column(sa.Integer, primary_key=True, autoincrement=True)
  coverage = sa.Column(sa.Float)
  completeness = sa.Column(sa.Float)

  # These column maps coverage/completeness to an individual+group
  sample_id = sa.Column(sa.String)
  group_id = sa.Column(sa.Integer)

  # Genetic relationship
  element_id = sa.Column(sa.String, sa.ForeignKey("Exon.id"))
  element = relationship("Exon", backref=backref("data"))

  def __init__(self, element_id, sample_id=None, group_id=None,
               coverage=None, completeness=None):
    super(ExonData, self).__init__()
    
    self.element_id = element_id
    self.sample_id = sample_id
    self.group_id = group_id
    self.coverage = coverage
    self.completeness = completeness


# =========================================================
#   The actual Element Adapter
# ---------------------------------------------------------
class ElementAdapter(ElementalDB):
  """
  SQLAlchemy-based :class:`ElementAdapter` for Chanjo.

  .. note::

    For testing pourposes; use ":memory:" as the `path` argument to set up
    in-memory version of the database.

  :param str path: Path to the database to connect to
  :param bool debug: Whether to print logging information (optional)
  """
  def __init__(self, path, debug=False):
    super(ElementAdapter, self).__init__(path, debug=debug)

    # Add new classes to the supported ORM classes
    self.classes.update({
      "gene_data": GeneData,
      "transcript_data": TranscriptData,
      "exon_data": ExonData
    })

  def transcriptStats(self, sample_id):
    """
    What's happening is that we are summing read depths and passed bases for
    each exon in a transcript and then dividing those numbers by the total
    exon length of the transcript.

    .. note::
      This needs to be carried out before annotating genes!
    """
    # I might turn this into a proper SQLAlchemy verison but for now.
    rawSQL = """
    SELECT Exon_Transcript.transcript_id,
           sum((Exon.end-Exon.start+1) * ExonData.coverage) / sum(Exon.end-Exon.start+1) AS coverage,
           sum((Exon.end-Exon.start+1) * ExonData.completeness) / sum(Exon.end-Exon.start+1) AS completeness
    FROM Exon
    INNER JOIN Exon_Transcript ON Exon.id=Exon_Transcript.exon_id
    INNER JOIN ExonData ON Exon.id=ExonData.element_id
    WHERE ExonData.sample_id="{sample_id}"
    GROUP BY Exon_Transcript.transcript_id
    """.format(sample_id=sample_id)

    return self.session.execute(rawSQL).fetchall()

  def geneStats(self, sample_id):
    """
    .. note::
      Annotation of transcripts needs to be acomplished before annotating genes!
    """
    rawSQL = """
    SELECT Gene.id,
           avg(TranscriptData.coverage),
           avg(TranscriptData.completeness)
    FROM Transcript
    INNER JOIN Gene ON Transcript.gene_id=Gene.id
    INNER JOIN TranscriptData ON Transcript.id=TranscriptData.element_id
    WHERE TranscriptData.sample_id="{sample_id}"
    GROUP BY Transcript.gene_id
    """.format(sample_id=sample_id)

    return self.session.execute(rawSQL).fetchall()



from __future__ import print_function

# Relative import shortcut
from core import Hub


__title__ = "chanjo"
__version__ = "0.2.0"
__author__ = "Robin Andeer"
__license__ = "MIT"
__copyright__ = "Copyright 2013 Robin Andeer"
__docformat__ = "restructuredtext"



#!/usr/bin/env python
# coding: utf-8
"""
  chanjo.bam
  ~~~~~~~~~~~~~

  The default :class:`CoverageAdapter` that ships with Chanjo. Talks directly
  to a BAM alignment file to extract read depth data.

  :copyright: (c) 2013 by Robin Andeer
  :license: MIT, see LICENSE for more details
"""

import pysam
import numpy as np


class CoverageAdapter(pysam.Samfile):
  """
  Adapter for interfacing directly with BAM alignment files.

  Usage::

    from chanjo.bam import CoverageAdapter
    path = "/path/to/bam/file.bam"
    adapter = CoverageAdapter(path)

  :param str bamPath: Path to the BAM alignment file
  """

  def __init__(self, bamPath):
    super(CoverageAdapter, self).__init__(bamPath, "rb")

  def read(self, chrom, start, end):
    """
    Public: Generates a list of read depths, for each position, between start,
    end.

    .. note::
      Positions are 0,0-based throughout Chanjo. If start=0, end=9 you should
      expect the 10 read depths for position 1-10 to be returned.

    Usage::

      adapter.read("17", 0, 5)
      #=> array([3., 3., 4., 4., 5., 4.])

    :param str chrom: The chromosome ID (str) of interest
    :param int start: The first position of the interval (0-based)
    :param int end: The last position of the interval (0-based)
    :returns: A numpy array of read depths for *each* position in the interval
    """
    # Generate a list of 0 read depth for each position
    positions = np.zeros(end+1-start)

    # Start Pileup iterator and walk through each position in the interval
    # `truncate` will make sure it starts and ends on the given positions!
    # +1 to end because pysam otherwise stops one base short by default
    for col in self.pileup(str(chrom), start, end+1, truncate=True):

      # Overwrite the read depth in the correct position
      # This will allow simple slicing to get at the positions of interest
      positions[col.pos - start] = col.n

    return positions



#!/usr/bin/env python
# coding: utf-8
"""
  chanjo.ccds2sql
  ~~~~~~~~~~~~~~~~

  Official importer module for the SQL :class:`ElementAdapter`. Works off of a
  CCDS database dump and extracts all the nessesary data to build up a complete
  genomic database in SQLite.

  :copyright: (c) 2013 by Robin Andeer
  :license: MIT, see LICENSE for more details
"""
import csv
from sql import ElementAdapter


class Importer(object):
  """
  Setup class that takes a CCDS database file and pushes it through to form the
  basis of a SQLite database. Should take only a few minutes to complete.

  :param str dbPath: Full path to the new database
  :param str ccdsPath: Path to location of the CCDS file (Default: ``None``,
                       optional)
  :param bool debug: Debug info is printed to the console (Def. `False`,
                     optional)
  """
  def __init__(self, dbPath, ccdsPath=None, debug=False):
    super(Importer, self).__init__()
    self.dbPath = dbPath
    self.reference(ccdsPath)

    # This is where the elements are stored before begin persisted
    self.genes = {}
    self.txs = {}
    self.exons = {}

    # Set up the Element Adapter
    self.adapter = ElementAdapter(dbPath, debug=debug)

  def reference(self, ccdsPath):
    """
    Public: Sets the path to the CCDS reference file

    :param str ccdsPath: Path to the CCDS file
    :returns: ``self`` for chainability
    """
    self.ccdsPath = ccdsPath

    return self

  def populate(self):
    """
    Public: Where it all happens. Populates the new database with data from the
    CCDS file.

    I think this is "ACID compliant". It only sets up things if everything
    works. I other words. If you have a working database after running this
    method you can be sure all elements were successfully added to the
    database.
    """
    # Setup the new database
    self.adapter.setup()

    self.reader = csv.reader(open(self.ccdsPath, "r"), delimiter="\t")

    for row in self.reader:

      # Skip some rows that don't contain useful information
      commentLine = row[0].startswith("#")
      unofficial = (row[5] != "Public")
      if commentLine or unofficial:
        continue

      (chrom, hgnc, tx_id, strand, tx_start,
       tx_end, exonCoords) = self._extractLineData(row)

      # Collect all transcript related information
      txKwargs = {
        "tx_id": tx_id,
        "chrom": chrom,
        "start": tx_start,
        "end": tx_end,
        "strand": strand,
        "gene_id": hgnc
      }

      # Check if the gene has already been created
      gene = self.genes.get(hgnc, None)
      if gene is None:
        # Collect all gene related information
        gsKwargs = {
          "hgnc": hgnc,
          "chrom": chrom,
          "start": tx_start,
          "end": tx_end,
          "strand": strand
        }

        gene = self.genes[hgnc] = self.adapter.create("gene", **gsKwargs)

      else:
        # Make sure we assign the correct gene coordinates (most extreme)
        if gene.start > tx_start:
          gene.start = tx_start

        if gene.end < tx_end:
          gene.end = tx_end

      tx = self.txs.get(tx_id, None)
      if tx is None:
        # Create a new transcript object
        tx = self.txs[tx_id] = self.adapter.create("transcript", **txKwargs)
        # Add the parent gene to the transcript
        tx.gene = gene

      # Create new exons
      for coord in exonCoords:
        ex_id = "{chrom}-{start}-{end}".format(chrom=chrom, start=coord[0],
                                               end=coord[1])

        if ex_id in self.exons:
          exon = self.exons[ex_id]

        else:
          # Collect all exon information
          exKwargs = {
            "ex_id": ex_id,
            "chrom": chrom,
            "start": coord[0],
            "end": coord[1],
            "strand": strand
          }
          exon = self.exons[ex_id] = self.adapter.create("exon", **exKwargs)

        # Add the gene and transcript parents to the exon
        if not gene in exon.genes:
          exon.genes.extend([gene])

        if not tx in exon.transcripts:
          exon.transcripts.extend([tx])

    # Add elements to session
    self.adapter.add(self.exons.values())
    self.adapter.add(self.genes.values())
    self.adapter.add(self.txs.values())

    self.adapter.commit()

  def _extractLineData(self, row):
    """
    Private: Extracts the useful information from one CCDS file row.

    :param list row: A list of strings from split row
    :returns: A bunch of strings and ints (see order below)
    """
    # Chrom, HGNC, Transcript ID, strand, Transcript start, end, exon coords
    return (row[0], row[2], row[4], row[6], int(row[7])-1, int(row[8])-1,
            self._generateExonCoordinates(row[9]))

  def _generateExonCoordinates(self, rowData):
    """
    Private: Takes the formatted string of exon coordinates from the CCDS row
    and turns it into a more managable list of lists with int start, end
    coordinates for each exon.

    :param str rowData: A csv string of (start,end) pairs
    :returns: A list of lists with the start, end pairs (int)
    """
    # Remove the "[]"
    csvExons = rowData[1:-1].replace(" ", "")

    # 1. Split first into exons coordinates
    # 2. Split into start, end and parse int
    exons = [[int(pos) for pos in item.split("-")]
             for item in csvExons.split(",")]

    # 3. Correct coords to 0,0-based Pythonic standard
    for exon in exons:
      exon[0] -= 1
      exon[1] -= 1

    return exons



#!/usr/bin/env python
# coding: utf-8
"""
  chanjo.core
  ~~~~~~~~~~~~~

  This module implements the glue and controlling hub. The :class:`Hub` is
  also where the adapters are plugged in to. This will likely be pretty much
  the only part of the package you directly interact with.

  :copyright: (c) 2013 by Robin Andeer
  :license: MIT, see LICENSE for more details
"""

import zlib
import itertools


class Hub(object):
  """
  The :class:`Hub` is the core component of Chanjo. When you plug in the
  adapters, it will create handy shortcuts to easily reach common methods.

  The :class:`Hub` is able to prepare data and annotate elements, calculate
  coverage for a range of genomic positions, and interact with the
  element data store.

  Usage::

    from chanjo.core import Hub
    from chanjo.bam import CoverageAdapter
    from chanjo.sqlite import ElementAdapter

    hub = Hub()
    bam_path = "/path/to/file.bam"
    cov_path = "/path/to/sqlite.db"
    hub.connect(CoverageAdapter(bam_path), ElementAdapter(cov_path))

  :param str coverageAdapter: (optional) Plug in the adapter during init
  :param str elementAdapter: (optional) Plug in the adapter during init

  """
  def __init__(self, coverageAdapter=None, elementAdapter=None):
    super(Hub, self).__init__()

    # Set up the adapters
    if coverageAdapter and elementAdapter:
      self.connect(coverageAdapter, elementAdapter)

  def connect(self, coverageAdapter, elementAdapter):
    """
    Public: Plugs in the required adapters and sets up a few shortcuts.

    Usage::

      from chanjo.core import Hub
      from chanjo.bam import CoverageAdapter
      from chanjo.sqlite import ElementAdapter

      hub = Hub()
      bam_path = "/path/to/file.bam"
      cov_path = "/path/to/sqlite.db"
      hub.connect(CoverageAdapter(bam_path), ElementAdapter(cov_path))

    :param str coverageAdapter: An instance of a :class:`CoverageAdapter`
    :param str elementAdapter:  An instance of a :class:`ElementAdapter`
    """
    # Customizable adapters
    self.cov = coverageAdapter
    self.db = elementAdapter

  def annotate(self, element, cutoff=10):
    """
    Public: Annotates each related exon with coverage data.

    Useage::

      genes = hub.db.get("gene", ["GIT1", "EGFR", "BRCA1"])
      for gene in genes:
        hub.annotate(gene, 15)

    :param object element: One element object (gene/transcript)
    :param cutoff: (optional) The min read depth to use for completeness
                   (Default: 10)
    """
    # Both transcripts and genes can be used to select exons to annotate
    depth = self.cov.read(element.chrom, element.start, element.end)

    # Preallocate list for each exon of the element
    exons = [None]*len(element.exons)

    # Get the exons related to the element
    for i, exon in enumerate(element.exons):
      # Relative start and end positions to slice the ``depth`` array
      start = exon.start - element.start
      end = exon.end - element.start

      # Do the heavy lifting
      # +1 to end because ``end`` is 0-based and slicing is 0,1-based
      (coverage, completeness,
       levels) = self.calculate(depth[start:end+1], cutoff)

      exons[i] = {
        "element_id": exon.id,
        "coverage": coverage,
        "completeness": completeness
      }

    return exons

  def calculate(self, depths, cutoff, levels=False):
    """
    Public: Calculates both coverage and completeness for a interval.

    Usage::

      gene = hub.db.get("gene", "C3")
      hub.coverage(gene.chrom, gene.intervals, 15)
      #=> (13.43522398231, 0.434122133123, None)

    :param list depths: List/array of the read depth for each position/base
    :param int cutoff: The cutoff for lowest passable read depth (completeness)
    :returns: Coverage (float), completeness (float), compressed levels (str)
    """
    # Initialize
    totBaseCount = float(len(depths))
    readCount = 0
    passedCount = 0

    for depth in depths:
      # Add the number of overlapping reads
      readCount += depth

      # Add the position if it passes `cutoff`
      if depth >= cutoff:
        passedCount += 1

    if levels:
      # Stringify and compress the levels to enable storage in SQL database
      str_levels = self.stringify(depths)
    else:
      str_levels = None

    # totBaseCount should never be able to be 0! Exons be >= 1 bp long
    return readCount / totBaseCount, passedCount / totBaseCount, str_levels

  def stringify(self, depths):
    """
    Public: Compresses the string of read depths.

    Because of how compression works I believe this will be sort of the same
    as generating BEDGraph intervals to compress the information.

    :param list depths: Array of read depth (float or int)
    :returns: Compressed string of read depths
    """
    # Turn floats to ints, then to strings, then concat with "|"-separator
    str_depths = "|".join(itertools.imap(str, map(int, depths)))

    # Compress and return compressed string
    return zlib.compress(str_depths)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import chanjo

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

setup(
    name="chanjo",
    description="A coverage analysis package for clinical sequencing.",
    author="Robin Andeer",
    author_email="robin.andeer@gmail.com",
    license="MIT",
    maintainer="Robin Andeer",
    keywords=["coverage", "sequencing", "clinical", "exome"],
    platform="UNIX",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 2.7",
        "Topic :: Scientific/Engineering :: Bio-Informatics"
    ],
    url="http://chanjo.readthedocs.org/",
    download_url="https://github.com/robinandeer/chanjo",
    version=chanjo.__version__,
    long_description="",
    install_requires=[
        "nose",
        "pysam",
        "numpy",
        "sqlalchemy",
        "interval",
        "docopt",
        "elemental",
        "path.py"
    ],
    packages=["chanjo"],
    scripts=["scripts/chanjo-autopilot"]
)